/*
You may use this source code for personal entertainment purposes only. Any commercial-, education- or military use is strictly forbidden without permission from the author. Any programs compiled from this source and the source code itself should be made available free of charge. Any modified versions must have the source code available for free upon request.
*/

//
// by Kegetys <kegetys@dnainternet.net>
//

// Script handling functions

#define WIN32_LEAN_AND_MEAN

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <windows.h>
#include <shellapi.h>
#include "testdll.h"
#include <ctype.h>

#define SCRIPT_VERSION	1.0			// Script engine version
#define MAX_PARAMS		6			// Maximum parameters for commands
#define WGET_MINWAIT	1000
#define PIPEBUFSIZE		4096
#define OFP_WINDOW_NAME "Operation Flashpoint"
int lastWget;

typedef struct {
	int	id;
	char cmd[128];
} COMMANDS;

// Valid commands
static enum {
	C_INFO_VERSION,
	C_INFO_DEBUGOUT,
	C_STRING_FIRST,
	C_STRING_LAST,
	C_STRING_INDEX,
	C_STRING_LENGTH,
	C_STRING_RANGE,
	C_STRING_TOUPPER,
	C_STRING_TOLOWER,
	C_STRING_TOARRAY,
	C_FILE_EXISTS,
	C_FILE_READ,
	C_FILE_WRITE,
	C_FILE_QWRITE,
	C_FILE_AWRITE,
	C_FILE_VARS,
	C_FILE_READVARS,
	C_FILE_REMOVE,
	C_FILE_DELETE,
	C_FILE_WGET,
	C_INPUT_GETKEYS,
	C_INPUT_GETKEY,
	C_INPUT_AGETKEYS,
	C_INPUT_AGETKEY,
	C_INPUT_GETMOUSE,
	C_INPUT_SETMOUSE,
};

COMMANDS cmdsList[] = {
	C_INPUT_GETMOUSE, "input getmouse",
	C_INPUT_SETMOUSE, "input setmouse",
	C_INPUT_GETKEYS, "input getkeys",
	C_INPUT_GETKEY, "input getkey",
	C_INPUT_AGETKEYS, "input agetkeys",
	C_INPUT_AGETKEY, "input agetkey",
	C_FILE_QWRITE, "file qwrite",
	C_FILE_EXISTS, "file exists",
	C_FILE_READ, "file read",
	C_FILE_AWRITE, "file awrite",
	C_FILE_WRITE, "file write",
	C_FILE_VARS, "file vars",
	C_FILE_READVARS, "file readvars",
	C_FILE_REMOVE, "file remove",
	C_FILE_DELETE, "file delete",
	C_FILE_WGET, "file wget",
	C_STRING_TOUPPER, "string toupper",
	C_STRING_TOLOWER, "string tolower",
	C_STRING_TOARRAY, "string toarray",
	C_STRING_LENGTH, "string length",
	C_STRING_INDEX, "string index",
	C_STRING_RANGE, "string range",
	C_STRING_FIRST, "string first",
	C_STRING_LAST, "string last",
	C_INFO_VERSION, "info version",
	C_INFO_DEBUGOUT, "info debugout",
	-1, NULL,
};

char* strtok2(char* str);
int matchCmd(char *cmd);
void QWrite(char* str, HANDLE file);
char* stripq(char *str);

// Parse a script command
void ParseScript(char* com, FULLHANDLE file) {

	int l = strlen(com);
	char *str = new char[l+1];
	if(!str)
		return;

	memcpy(str, com, l+1);
	HANDLE out = file.handle;

	// Tokenize the string	
	str = strtok2(str);
	char *pch = strtok(str,"\n");

	char *par[MAX_PARAMS];
	int numP = 0;

	while (pch != NULL && numP < MAX_PARAMS) {
		par[numP] = pch;		
		pch = strtok(NULL, "\n");
		numP++;
	}

	// Find command
	char *temp = new char[l];
	sprintf(temp, "%s %s", par[0], par[1]);
	switch(matchCmd(temp)) {
		case C_INFO_VERSION:
			{ // Return version
				char ret[12];
				sprintf(ret, "%.2f", SCRIPT_VERSION);
				QWrite(ret, out);
			}
			break;

		case C_INFO_DEBUGOUT:
			{ // Output debug string
				if(numP < 3) {
					QWrite("ERROR: Not enough parameters", out);
					break;
				}

				char *str1 = stripq(par[2]);
				DebugMessage(str1);
				QWrite("1", out);
			}
			break;

		case C_STRING_FIRST:
			{ // Return the index in str2 of the first occurrence of str1, or -1 if str1 is not found.
				if(numP < 4) {
					QWrite("ERROR: Not enough parameters", out);
					break;
				}

				char *str1 = stripq(par[2]);
				char *str2 = stripq(par[3]);
				char *res = strstr(str2, str1);
				
				if(!res) {
					// It wasn't found
					QWrite("-1", out);
				} else {
					// Found, return relative pointer
					char ret[16];
					sprintf(ret, "%d", res - str2);
					QWrite(ret, out);
				}

			}
			break;

		case C_STRING_LAST:
			{ // Return the index in str2 of the last occurrence of str1, or -1 if str1 is not found.
				if(numP < 4) {
					QWrite("ERROR: Not enough parameters", out);
					break;
				}

				char *str1 = stripq(par[2]);
				char *str2 = stripq(par[3]);
				char *str2t = stripq(par[3]);

				char *res = 0, *res2 = 0;
				do {
					// Search substring until its not found anymore
					res = res2;
					res2 = strstr(str2t, str1);
					str2t = res2+1;
				} while (res2 != NULL);
				
				if(!res) {
					// It wasn't found at all
					QWrite("-1", out);
				} else {
					// Found, return relative pointer
					char ret[16];
					sprintf(ret, "%d", res - str2);
					QWrite(ret, out);
				}

			}
			break;

		case C_STRING_LENGTH:
			{ // Return the number of characters in string.
				if(numP < 3) {
					QWrite("ERROR: Not enough parameters", out);
					break;
				}

				char *str = stripq(par[2]);
				int res = strlen(str);
				char ret[16];
				sprintf(ret, "%d", res);
				QWrite(ret, out);

			}
			break;

		case C_STRING_RANGE:
			{ // Return the range of characters in str from i to j.
				if(numP < 5) {
					QWrite("ERROR: Not enough parameters", out);
					break;
				}

				char *str = stripq(par[2]);
				unsigned int l = strlen(str);
				unsigned int i = atoi(par[3]);
				unsigned int j = atoi(par[4]);

				if(i > l) i = l;
				if(j > l) j = l;

				str[j] = '\0';
				char *ret = new char[l+4];
				if(!ret)
					break;

				sprintf(ret, "\"%s\"", str+i);
				QWrite(ret, out);
				delete[] ret;

			}
			break;

		case C_STRING_TOLOWER:
			{ // Return string in lower case.
				if(numP < 3) {
					QWrite("ERROR: Not enough parameters", out);
					break;
				}

				char *str = stripq(par[2]);
				unsigned int l = strlen(str);

				for(unsigned int x=0;x < l;x++)
					str[x] = tolower(str[x]);

				char *ret = new char[l+4];
				if(!ret)
					break;

				sprintf(ret, "\"%s\"", str);
				QWrite(ret, out);
				delete[] ret;
			}
			break;

		case C_STRING_TOUPPER:
			{ // Return string in upper case.
				if(numP < 3) {
					QWrite("ERROR: Not enough parameters", out);
					break;
				}

				char *str = stripq(par[2]);
				unsigned int l = strlen(str);

				for(unsigned int x=0;x < l;x++)
					str[x] = toupper(str[x]);

				char *ret = new char[l+4];
				if(!ret)
					break;

				sprintf(ret, "\"%s\"", str);
				QWrite(ret, out);
				delete[] ret;
			}
			break;

		case C_STRING_TOARRAY:
			{ // Return all characters in string as array elements
				if(numP < 3) {
					QWrite("ERROR: Not enough parameters", out);
					break;
				}

				char *str = stripq(par[2]);
				unsigned int l = strlen(str);

				QWrite("[", out);
				for(unsigned int x=0;x < l;x++) {
					char ret[5];
					sprintf(ret, "\"%c\"", str[x]);
					QWrite(ret, out);
					if(x < l-1)
						QWrite(",", out);
				}
				QWrite("];", out);
			}
			break;

		case C_STRING_INDEX:
			{ // Return the character at the specified index.
				if(numP < 4) {
					QWrite("ERROR: Not enough parameters", out);
					break;
				}

				unsigned int idx = atoi(par[3]);
				par[2] = stripq(par[2]);

				if(idx > (strlen(par[2])-1) || idx < 0) 
					QWrite("ERROR: Index out of bounds", out);
				else {
					// Write char to file
					char c = par[2][idx];
					char ret[6];

					sprintf(ret, "\"%c\"", par[2][idx]);
					QWrite(ret, out);
				}
			}
			break;

		case C_FILE_EXISTS:
			{ // Return 1 if name exists, else 0.
				if(numP < 3) {
					QWrite("ERROR: Not enough parameters", out);
					break;
				}

				char *file = stripq(par[2]);
				if(fdbExists(file))
					QWrite("1", out);
				else
					QWrite("-1", out);
				
			}
			break;

		case C_FILE_READ:
			{ // Read val from file fwatcher/mdb/name
				if(numP < 4) {
					QWrite("ERROR: Not enough parameters", out);
					break;
				}

				char *file = stripq(par[2]);
				char *var = stripq(par[3]);
				QWrite(fdbGet(file, var), out);
			}
			break;
			
		case C_FILE_WRITE:
			{ // Write var=val to file fwatcher/mdb/name
				if(numP < 5) {
					QWrite("ERROR: Not enough parameters", out);
					break;
				}

				char *file = stripq(par[2]);
				char *var = stripq(par[3]);
				char *val = par[4];
				if(fdbPut(file, var, val, false))
					QWrite("1", out);
				else
					QWrite("-1", out);
			}
			break;

		case C_FILE_QWRITE:
			{ // Write var=val to file fwatcher/mdb/name without checking if it already exists
				if(numP < 5) {
					QWrite("ERROR: Not enough parameters", out);
					break;
				}

				char *file = stripq(par[2]);
				char *var = stripq(par[3]);
				char *val = par[4];
				if(fdbPutQ(file, var, val))
					QWrite("1", out);
				else
					QWrite("-1", out);
			}
			break;

		case C_FILE_AWRITE:
			{ // Append val to var in file fwatcher/mdb/name
				if(numP < 5) {
					QWrite("ERROR: Not enough parameters", out);
					break;
				}

				char *file = stripq(par[2]);
				char *var = stripq(par[3]);
				char *val = par[4];
				if(fdbPut(file, var, val, true))
					QWrite("1", out);
				else
					QWrite("-1", out);
			}
			break;

		case C_FILE_VARS:
			{ // Return array of vars in file fwatcher/mdb/name
				if(numP < 3) {
					QWrite("ERROR: Not enough parameters", out);
					break;
				}

				char *file = stripq(par[2]);
				char *res = fdbVars(file);
				QWrite(res, out);
				delete[] res;
			}
			break;

		case C_FILE_READVARS:
			{ // Return all vars from file fwatcher/mdb/name
				if(numP < 3) {
					QWrite("ERROR: Not enough parameters", out);
					break;
				}

				char *file = stripq(par[2]);
				char *res = fdbReadvars(file);
				QWrite(res, out);
				delete[] res;
			}
			break;

		case C_FILE_REMOVE:
			{ // Remove a val from file fwatcher/mdb/name
				if(numP < 4) {
					QWrite("ERROR: Not enough parameters", out);
					break;
				}

				char *file = stripq(par[2]);
				char *var = stripq(par[3]);
				if(fdbRemove(file, var))
					QWrite("1", out);
				else
					QWrite("-1", out);
			}
			break;

		case C_FILE_DELETE:
			{ // Remove file fwatcher/mdb/name
				if(numP < 3) {
					QWrite("ERROR: Not enough parameters", out);
					break;
				}

				char *file = stripq(par[2]);
				if(fdbDelete(file))
					QWrite("1", out);
				else
					QWrite("-1", out);
			}
			break;

		case C_FILE_WGET:
			{ // Get a http/ftp file with wget
				// Execute wget, outputing to a named pipe
				// Not very clean but it works...
				if(numP < 3) {
					QWrite("ERROR: Not enough parameters", out);
					break;
				}

				if((GetTickCount() - lastWget) < WGET_MINWAIT) {
					// Return -1 if user is calling wget too often
					QWrite("-1", out);
					break;
				}

				// wget should always return an empty file if an error occurs so no handling should be required for it
				// TODO: maybe should return some sort of error message for OFP...

				HANDLE pipe;
				if(nomap)
					// Create a named pipe for wget to write into					
					pipe = CreateNamedPipe(file.filename, PIPE_ACCESS_DUPLEX, PIPE_TYPE_BYTE | PIPE_READMODE_BYTE | PIPE_WAIT,
											PIPE_UNLIMITED_INSTANCES, PIPEBUFSIZE, PIPEBUFSIZE, NMPWAIT_USE_DEFAULT_WAIT, NULL);

				char *cline = new char[256+strlen(file.filename)+strlen(par[2])];
				sprintf(cline, "-q --tries=1 --output-document=\"%s\" --timeout=3 --user-agent=fwatch/%.2f %s", file.filename, SCRIPT_VERSION, par[2]);

				// Win23 API bloat
				STARTUPINFO si;
				PROCESS_INFORMATION pi;
				ZeroMemory(&si, sizeof(si));
				si.cb = sizeof(si);
				si.dwFlags = STARTF_USESHOWWINDOW;
				si.wShowWindow = SW_HIDE;
				ZeroMemory(&pi, sizeof(pi));

				// Execute wget
				CreateProcess("fwatch/data/wget.exe", cline, NULL, NULL, TRUE, HIGH_PRIORITY_CLASS, NULL, NULL, &si, &pi);
				delete[] cline;
				lastWget = GetTickCount();
				
				if(nomap) {
					// Wait for wget to do its thing
					char *buf = new char[PIPEBUFSIZE+1];
					if(!buf)
						break;

					DWORD st, foo = 1, foo2;
					do {					
						// Check exit code
						GetExitCodeProcess(pi.hProcess, &st);
						Sleep(5); // Lets not eat all cpu cycles

						// Read named pipe and output to the out pipe for OFP
						ReadFile(pipe, buf, PIPEBUFSIZE, &foo, NULL);
						//buf[foo] = 0x00;
						if(foo)
							WriteFile(out, buf, foo, &foo2, NULL);

						// Wait until the process has exited and no more data in pipe
					} while(st == STILL_ACTIVE || foo);

					WriteFile(out, "\0", 1, &foo2, NULL);
					CloseHandle(pipe); 
					delete[] buf;
				} else {
					DWORD st;
					do {					
						// Check exit code
						GetExitCodeProcess(pi.hProcess, &st);
						Sleep(10); // Lets not eat all cpu cycles
						// Wait until the process has exited
					} while(st == STILL_ACTIVE);
				}
			}
			break;

		case C_INPUT_GETKEYS:
			{ // Return array of all keys pressed
				// TODO: skip all input things on dedicated server
				if(numP < 2) {
					QWrite("ERROR: Not enough parameters", out);
					break;
				}

				QWrite("[", out);				
				bool f = false;
				if(checkActiveWindow())
					for(int i=0;i < 128;i++) 
						if((GetKeyState(i) & 0x8000) >> 15) {
							if(f)
								QWrite(",", out);
							QWrite(formatKey(i), out);
							f = true;
						}
				QWrite("]", out);
			}
			break;

		case C_INPUT_GETKEY:
			{ // Return 1 if key pressed
				if(numP < 3) {
					QWrite("ERROR: Not enough parameters", out);
					break;
				}
				char idx = toupper(par[2][0]);
				
				if((GetKeyState(idx) & 0x8000) >> 15 && checkActiveWindow())
					QWrite("1", out);
				else
					QWrite("-1", out);
			}
			break;

		case C_INPUT_AGETKEYS:
			{ // Return array of all keys pressed, works asynchronously
				if(numP < 2) {
					QWrite("ERROR: Not enough parameters", out);
					break;
				}

				QWrite("[", out);
				bool f = false;
				if(checkActiveWindow())
					for(int i=0;i < 128;i++) 
						if(GetAsyncKeyState(i) & 0x0001) {
							if(f)
								QWrite(",", out);
							QWrite(formatKey(i), out);
							f = true;
						}
				QWrite("]", out);
			}
			break;

		case C_INPUT_AGETKEY:
			{ // Return 1 if key pressed, works asynchronously
				if(numP < 3) {
					QWrite("ERROR: Not enough parameters", out);
					break;
				}
				char idx = toupper(par[2][0]);
				
				if(GetAsyncKeyState(idx) & 0x0001 && checkActiveWindow())
					QWrite("1", out);
				else
					QWrite("-1", out);
			}
			break;

		case C_INPUT_GETMOUSE:
			{ // Get mouse data [x, y, leftmouse, rightmouse, middlemouse]
				if(numP < 2) {
					QWrite("ERROR: Not enough parameters", out);
					break;
				}
				char ret[256];
				
				// Get mouse position etc.
				POINT mp;
				GetCursorPos(&mp);
				sprintf(ret, "[%d, %d, %s, %s, %s]", mp.x, mp.y, getBool(GetAsyncKeyState(VK_LBUTTON)), 
													 getBool(GetAsyncKeyState(VK_RBUTTON)), getBool(GetAsyncKeyState(VK_MBUTTON)));
				QWrite(ret, out);
			}
			break;

		case C_INPUT_SETMOUSE:
			{ // Set mouse cursor position
				if(numP < 4) {
					QWrite("ERROR: Not enough parameters", out);
					break;
				}

				unsigned int mx = atoi(par[2]);
				unsigned int my = atoi(par[3]);
				
				// Get mouse position etc.
				if(checkActiveWindow()) { // Do not set cursor pos is OFP window is not active
					if(SetCursorPos(mx, my))
						QWrite("1", out);
					else
						QWrite("-1", out);
				} else
					QWrite("-1", out);
			}
			break;

		default:
			QWrite("ERROR: Unknown command", out);
			break;
	}

	delete[] temp;
	delete[] str;
	return;
}

// Remove quotation marks from string
char* stripq(char *str) {
	if(str[0] == '"' && str[strlen(str)-1] == '"') {
		str++;
		str[strlen(str)-1] = '\0';
	}
	return str;
}

// "Tokenize" string, ignoring substrings (inside quotation marks)
// Ugly
char* strtok2(char* str) {
	char c;
	int i = 0, y = 0;
	bool x = false;

	// Replace spaces with null character
	do {
		c = str[i];

		if(c == '"')
			x=!x;

		if(c == ' ' && !x)
			str[i] = '\n';
		i++;
	} while(c != '\0');
	
	i = 0, y = 0;
	// Remove double-null characters
	do {
		str[y] = str[i];
		if(str[i] != '\0')
			if(str[i] != '\n' || str[i+1] != '\n')
				y++;
		i++;
	} while(str[y] != '\0');

	return str;
}

// Find matching command and return ID or -1 if not found
int matchCmd(char *cmd) {
	int cc = 0, i = 0;
	do {
		cc = cmdsList[i].id;
		if(!strcmp(cmd, (const char *) &cmdsList[i].cmd))
			return cc;
		i++;
	} while(cc != -1);
	return -1;
}

void QWrite(char* str, HANDLE file) {
	DWORD foo;
	unsigned int l = strlen(str);
	WriteFile(file, str, l, &foo, NULL);

	if(foo < l)
		DebugMessage("WARNING! Wanted to write %d bytes but only %d written", strlen(str), foo);
}

// Return "TRUE" or "FALSE"
inline char* getBool(short b) {
	if(b != 0)
		return "TRUE";
	else
		return "FALSE";
}

// Format special keys to string
char *formatKey(int c) {
	switch(c) {
		case VK_F1: return "\"F1\"";
		case VK_F2: return "\"F2\"";
		case VK_F3: return "\"F3\"";
		case VK_F4: return "\"F4\"";
		case VK_F5: return "\"F5\"";
		case VK_F6: return "\"F6\"";
		case VK_F7: return "\"F7\"";
		case VK_F8: return "\"F8\"";
		case VK_F9: return "\"F9\"";
		case VK_F10: return "\"F10\"";
		case VK_F11: return "\"F11\"";
		case VK_F12: return "\"F12\"";
		case VK_TAB: return "\"TAB\"";
		case VK_ESCAPE: return "\"ESC\"";
		case VK_SPACE: return "\"SPACE\"";
		case VK_SHIFT: return "\"SHIFT\"";
		case VK_CONTROL: return "\"CTRL\"";
		case VK_MENU: return "\"ALT\"";
		case VK_CAPITAL: return "\"CAPSLOCK\"";
		case VK_BACK: return "\"BACKSPACE\"";
		case VK_RETURN: return "\"ENTER\"";
		case VK_PRIOR: return "\"PAGEUP\"";
		case VK_NEXT: return "\"PAGEDOWN\"";
		case VK_END: return "\"END\"";
		case VK_HOME: return "\"HOME\"";
		case VK_LEFT: return "\"LEFT\"";
		case VK_RIGHT: return "\"RIGHT\"";
		case VK_UP: return "\"UP\"";
		case VK_DOWN: return "\"DOWN\"";
		case VK_EXECUTE: return "\"EXECUTE\"";
		case VK_INSERT: return "\"INSERT\"";
		case VK_DELETE: return "\"DELETE\"";
		case VK_NUMLOCK: return "\"NUMLOCK\"";
		case VK_SCROLL: return "\"SCROLLOCK\"";
		case VK_SELECT: return "\"SELECT\"";
		case VK_MULTIPLY: return "\"SELECT\"";
		case VK_ADD: return "\"ADD\"";
		case VK_SUBTRACT: return "\"SUBTRACT\"";
		case VK_DECIMAL: return "\"DECIMAL\"";
		case VK_DIVIDE: return "\"DIVIDE\"";
		case VK_NUMPAD0: return "\"NUMPAD0\"";
		case VK_NUMPAD1: return "\"NUMPAD1\"";
		case VK_NUMPAD2: return "\"NUMPAD2\"";
		case VK_NUMPAD3: return "\"NUMPAD3\"";
		case VK_NUMPAD4: return "\"NUMPAD4\"";
		case VK_NUMPAD5: return "\"NUMPAD5\"";
		case VK_NUMPAD6: return "\"NUMPAD6\"";
		case VK_NUMPAD7: return "\"NUMPAD7\"";
		case VK_NUMPAD8: return "\"NUMPAD8\"";
		case VK_NUMPAD9: return "\"NUMPAD9\"";
				
		default:
			static char s[5];
			sprintf(s, "\"%c\"", c);
			return s;
	}
}

// Return true if OFP is the active window (OFP client only, returns false on dedicated server)
bool checkActiveWindow(void) {
	char wn[128];
	GetWindowText(GetForegroundWindow(), wn, 128);
	if(!strcmp(wn, OFP_WINDOW_NAME))
		return true;
	else
		return false;
}